/*
 * TCS Dependency Map Renderer
 *
 * Copyright (c) 2005 TopCoder, Inc. All rights reserved.
 *
 * Author: Pops, TCSDeveloper
 */

using System;
using System.Collections;
using System.Drawing;
using System.Drawing.Drawing2D;

namespace TopCoder.Rules.DependencyMap.View
{

    /// <summary>
    /// This class is an implementation of the ILayoutManager that will layout the dependency diagram into a fixed
    /// sized grid (where all columns are fixed to a single width and all rows are fixed to a single height).
    /// </summary>
    public class FixedGridLayoutManager : ILayoutManager
    {

        /// <summary>
        /// This variable represents the extra padding that should appear between columns. This variable is only
        /// used *between* columns and will not provide margins around the outer columns. This variable is set in
        /// the constructor, is immutable once set and will be used by the layout mechanisms.
        /// </summary>
        private readonly int colPadding;

        /// <summary>
        /// This variable represents the extra padding that should appear between rows. This variable is only used
        /// *between* rows and will not provide margins around the outer rows. This variable is set in the
        /// constructor, is immutable once set and will be used by the layout mechanisms.
        /// </summary>
        private readonly int rowPadding;

        /// <summary>
        /// This variable represents the margins that should be maintained between the text and shape. This variable
        /// is set in the constructor, is immutable once set and will be used by the layout mechanisms.
        /// </summary>
        private readonly int textMargin;

        /// <summary>
        /// The font to use.
        /// </summary>
        private readonly Font font;

        /// <summary>
        /// Constructs the fixed grid layout manager using a 10 pixel margin between cols and rows. This constructor
        /// simply calls "this(10,10)".
        /// </summary>
        public FixedGridLayoutManager() : this(10, 10)
        {
        }

        /// <summary>
        /// Constructs the fixed grid layout using the specified row/col padding with a default text margin of 10
        /// pixels around all edges. This constructor should simply execute 'this(colPadding, rowPadding, 10)'.
        /// </summary>
        /// <param name="colPadding">The padding between columns</param>
        /// <param name="rowPadding">The padding between rows</param>
        /// <exception cref="ArgumentOutOfRangeException">if either colPadding or rowPadding is
        /// less than 0</exception>
        public FixedGridLayoutManager(int colPadding, int rowPadding) : this(colPadding, rowPadding, 10)
        {
        }

        /// <summary>
        /// Constructs the fixed grid layout manager using the specified padding for cols/rows and margin around
        /// text. This constructor simply saves the parameters to the like named variables.
        /// </summary>
        /// <param name="colPadding">The padding between columns</param>
        /// <param name="rowPadding">The padding between rows</param>
        /// <param name="textMargin">The margin between the text and the shape</param>
        /// <exception cref="ArgumentOutOfRangeException">if either colPadding or rowPadding or textMargin is
        /// less than 0</exception>
        public FixedGridLayoutManager(int colPadding, int rowPadding, int textMargin)
        {
            if (colPadding < 0)
            {
                throw new ArgumentOutOfRangeException("colPadding should not be negative.");
            }
            if (rowPadding < 0)
            {
                throw new ArgumentOutOfRangeException("rowPadding should not be negative.");
            }
            if (textMargin < 0)
            {
                throw new ArgumentOutOfRangeException("textMargin should not be negative.");
            }
            this.colPadding = colPadding;
            this.rowPadding = rowPadding;
            this.textMargin = textMargin;

            this.font = new Font(FontFamily.GenericSansSerif, 12);
        }

        /// <summary>
        /// Calculate the widths of each of the unseen dependencies with is referenced by a root dependency.
        /// </summary>
        /// <param name="dependency">the root dependency</param>
        /// <returns>a mapping from dependency to widths (Integer)</returns>
        private IDictionary GetWidths(Dependency dependency)
        {
            // Mapping from dependants to widths (number of cells)
            IDictionary widths = new Hashtable();
            // Visitation count for dependants (int)
            IDictionary seen = new Hashtable();
            // BFS queue
            IList queue = new ArrayList();

            // Initialize the queue and visitation count by placing the root dependency.
            queue.Add(dependency);
            seen[dependency] = 1;

            // Iterate through the queue.
            for (int i = 0; i < queue.Count; ++i)
            {
                dependency = queue[i] as Dependency;
                foreach (Dependency dependant in dependency.Dependants)
                {

                    // If the dependency is seen for the first time, we need to calculate its
                    // width, otherwise it's a seen dependency and has constant width of 1.
                    if (!seen.Contains(dependant))
                    {
                        queue.Add(dependant);
                        seen[dependant] = 1;
                    }
                    else
                    {
                        seen[dependant] = ((int) seen[dependant]) + 1;
                    }
                }
            }

            // Calculate the width in a backward order.
            for (int i = queue.Count - 1; i >= 0; --i) {
                dependency = queue[i] as Dependency;

                int width = 0;
                foreach (Dependency dependant in dependency.Dependants)
                {

                    // If the dependency is an unseen dependency, its width should already been calculated.
                    // Otherwise its width is constantly 1.
                    if ((int) seen[dependant] == 1)
                    {
                        width += (int) widths[dependant];
                    }
                    else
                    {
                        ++width;
                        seen[dependant] = ((int) seen[dependant]) - 1;
                    }
                }

                // If width is even make it odd.
                if (width % 2 == 0) {
                    ++width;
                }

                // Store the width into the mapping.
                widths[dependency] = width;
            }

            return widths;
        }

        /// <summary>
        /// Paint a dependency.
        /// </summary>
        /// <param name="dependency">the dependency to paint</param>
        /// <param name="shape">the shape for outbound</param>
        /// <param name="cellSize">the size of the cell</param>
        /// <param name="arrowHeight">the height of the arrow</param>
        /// <param name="x">the x coord position</param>
        /// <param name="y">the y coord position</param>
        /// <param name="graphics">the graphics to paint onto</param>
        private void PaintDependency(Dependency dependency, GraphicsPath shape, SizeF cellSize,
                                     float arrowHeight, float x, float y, IGraphics graphics) {
            // Calculate pixel offset.
            x *= cellSize.Width + this.colPadding;
            y *= cellSize.Height + arrowHeight + this.rowPadding;

            RectangleF rect = shape.GetBounds();
            // Scale the shape to cell size.
            Matrix matrix = new Matrix();
            matrix.Scale(cellSize.Width / rect.Width, cellSize.Height / rect.Height);
            shape.Transform(matrix);

            // Translate the shape to position.
            matrix = new Matrix();
            matrix.Translate(x, y);
            shape.Transform(matrix);

            // Draw the shape.
            graphics.DrawPath(new Pen(Color.Black), shape);

            // Estimate text width.
            SizeF myTextSize = graphics.MeasureString(dependency.Title, this.font);

            // Draw the text to the center of the cell.
            PointF position = new PointF(x + (cellSize.Width - myTextSize.Width) / 2,
                                         y + (cellSize.Height + myTextSize.Height) / 2);
            graphics.DrawString(dependency.Title, font, new SolidBrush(Color.Black), position);
        }

        /// <summary>
        /// Paint an arrow.
        /// </summary>
        /// <param name="shape">the shape for the arrow</param>
        /// <param name="x1">the x coord of the first dependency</param>
        /// <param name="x2">the x coord of the second dependency</param>
        /// <param name="y">the y coord of the first dependency</param>
        /// <param name="cellSize">the size of a cell</param>
        /// <param name="graphics">the graphics to paint onto</param>
        private void PaintArrow(GraphicsPath shape, float x1, float x2, float y, SizeF cellSize, IGraphics graphics) {
            RectangleF rect = shape.GetBounds();
            // Calculate arrow position.
            float x = (x1 + x2) / 2;
            x *= cellSize.Width + this.colPadding;
            x += (cellSize.Width - rect.Width) / 2;
            y *= cellSize.Height + rect.Height + this.rowPadding;
            y += cellSize.Height + this.rowPadding / 2;

            // Ratate the shape to point the dependant.
            Matrix matrix = new Matrix();
            matrix.RotateAt((float) (90 - 180 / Math.PI *
                    Math.Atan2(cellSize.Height + rect.Height + this.rowPadding,
                               (x1 - x2) * (cellSize.Width + this.colPadding))),
                            new PointF(rect.Width / 2, rect.Height / 2));
            shape.Transform(matrix);

            // Translate the shape to position.
            matrix = new Matrix();
            matrix.Translate(x, y);
            shape.Transform(matrix);

            // Draw the arrow.
            graphics.DrawPath(new Pen(Color.Black), shape);
        }

        /// <summary>
        /// Paint the dependency hiearchy.
        /// </summary>
        /// <param name="dependency">the root dependency</param>
        /// <param name="widths">the mapping of the width of the dependencies</param>
        /// <param name="shapeFactory">the shape factory to produce the shapes</param>
        /// <param name="graphics">the graphics to paint onto</param>
        private void Paint(Dependency dependency, IDictionary widths, ShapeFactory shapeFactory, IGraphics graphics) {
            // Calculate the maximum text size.
            SizeF textSize = new SizeF(0, 0);
            foreach (Dependency dependant in widths.Keys)
            {
                SizeF myTextSize = graphics.MeasureString(dependant.Title, this.font);
                if (myTextSize.Height > textSize.Height)
                {
                    textSize.Height = myTextSize.Height;
                }
                if (myTextSize.Width > textSize.Width)
                {
                    textSize.Width = myTextSize.Width;
                }
            }
            SizeF cellSize = new SizeF(textSize.Width + this.textMargin * 2 + 2,
                                       textSize.Height + this.textMargin * 2 + 2);
            float arrowHeight = shapeFactory[ShapeFactory.RelationshipArrow].GetBounds().Height;

            // Visitation set
            IDictionary seen = new Hashtable();
            // BFS queue
            IList queue = new ArrayList();
            // Position of the dependency to render
            IList positions = new ArrayList();
            // The position of the parent dependency
            IList parentPos = new ArrayList();

            // Initialize by inserting the root dependency.
            seen[dependency] = true;
            queue.Add(dependency);
            positions.Add(new PointF(0, 0));
            parentPos.Add(null);

            // Render through the queue.
            for (int i = 0; i < queue.Count; ++i)
            {
                dependency = queue[i] as Dependency;
                PointF position = (PointF) positions[i];

                // Calculate the dependants width to decide whether we need a padding cell.
                int dependantsWidth = 0;
                foreach (Dependency dependant in dependency.Dependants)
                {
                    if (!seen.Contains(dependant))
                    {
                        dependantsWidth += (int) widths[dependant];
                    }
                    else
                    {
                        ++dependantsWidth;
                    }
                }
                bool padding = dependantsWidth % 2 == 0;

                // If parent dependency exists we make the relationship arrow.
                if (parentPos[i] != null)
                {
                    float parent = (float) parentPos[i];
                    PaintArrow(shapeFactory[ShapeFactory.RelationshipArrow],
                               parent, position.X + dependantsWidth / 2,
                               position.Y - 1, cellSize, graphics);
                }

                // Obtain shape used for this unseen shape.
                GraphicsPath shape = null;
                if (dependency.ShapeOverride != -1)
                {
                    shape = shapeFactory[dependency.ShapeOverride];
                }
                if (shape == null)
                {
                    shape = shapeFactory[ShapeFactory.UnseenElement];
                }
                PaintDependency(dependency, shape, cellSize,
                                arrowHeight, position.X + dependantsWidth / 2, position.Y, graphics);

                // Accumulate the width to find the position for each dependant.
                int cumulativeWidth = 0;
                foreach (Dependency dependant in dependency.Dependants)
                {

                    // Check to see if we make the padding on this dependant.
                    if (padding && cumulativeWidth * 2 >= dependantsWidth)
                    {
                        padding = false;
                        ++cumulativeWidth;
                    }

                    if (!seen.Contains(dependant))
                    {

                        // Push the unseen dependant into the queue.
                        seen[dependant] = true;
                        queue.Add(dependant);
                        positions.Add(new PointF(position.X + cumulativeWidth, position.Y + 1));
                        parentPos.Add(position.X + dependantsWidth / 2);

                        cumulativeWidth += (int) widths[dependant];
                    }
                    else
                    {
                        // Draw arrow for the seen dependant.
                        PaintArrow(shapeFactory[ShapeFactory.RelationshipArrow],
                                   position.X + dependantsWidth / 2, position.X + cumulativeWidth,
                                   position.Y, cellSize, graphics);

                        // Obtain shape used for this seen shape.
                        shape = null;
                        if (dependant.ShapeOverride != -1)
                        {
                            shape = shapeFactory[dependant.ShapeOverride];
                        }
                        if (shape == null)
                        {
                            shape = shapeFactory[ShapeFactory.SeenElement];
                        }

                        // Draw the dependant.
                        PaintDependency(dependant, shape, cellSize,
                                        arrowHeight, position.X + cumulativeWidth, position.Y + 1, graphics);

                        ++cumulativeWidth;
                    }
                }
            }
        }

        /// <summary>
        /// Lays out and prints the dependency diagram. Since the algorithm to do this is beyond Poseidon's
        /// capabilities - please review the component specifications for details.
        /// </summary>
        /// <param name="dependency">The dependency to render</param>
        /// <param name="shapeFactory">The shape factory to use</param>
        /// <param name="graphics">The IGraphics to print to</param>
        /// <exception cref="ArgumentNullException">if either dependency, shapeFactory or graphics is null</exception>
        /// <exception cref="RendererException">if dependency contains non-Dependency dependencies
        /// (tongue twister eh)</exception>
        public void Layout(Dependency dependency, ShapeFactory shapeFactory, IGraphics graphics)
        {
            if (dependency == null)
            {
                throw new ArgumentNullException("dependency");
            }
            if (shapeFactory == null)
            {
                throw new ArgumentNullException("shapeFactory");
            }
            if (graphics == null)
            {
                throw new ArgumentNullException("graphics");
            }

            IDictionary widths = GetWidths(dependency);

            Paint(dependency, widths, shapeFactory, graphics);
        }

    }

}
