/*
 * TCS Dependency Map Renderer
 *
 * Copyright (c) 2005 TopCoder, Inc. All rights reserved.
 *
 * Author: Pops, TCSDeveloper
 */

using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using TopCoder.SVG.Generator;

namespace TopCoder.Rules.DependencyMap.View
{

    /// <summary>
    /// This class is an Adapter class between the IGraphics interface and the SVG Generator component
    /// (specifically the SVGDrawer class). This class attempts to simplify the SVG component and offer only
    /// those features the IGraphics interface requires. This class is created by the SVGRenderer in response
    /// to the GetGraphics call and will simply proxy its call to the embedded SVGDrawer class.
    /// Note: the SVGRenderer should never reuse this class because there is no way to reset the SVG information
    /// from an older drawing.
    /// </summary>
    internal class SVGGraphicsAdapter : IGraphics
    {

        /// <summary>
        /// This represents the underlying SVGDrawer that this adapter class uses. This variable is created at
        /// object instantiation time, is immutable (the reference, not the contents) and is accessed by all
        /// methods. This variable has been marked readonly to ensure it's immutability.
        /// </summary>
        private readonly SVGDrawer svgDrawer = new SVGDrawer();

        /// <summary>
        /// This variable represents the file name that the diagram will be rendered to. The variable is initially
        /// set in the constructor, is immutable and will never be null. The variable is accessed in the Complete
        /// method. This variable has been marked readonly to ensure it's immutability.
        /// </summary>
        private readonly string filename;

        /// <summary>
        /// Constructors the adapter class. Simple saves the argument to the like named variable.
        /// </summary>
        /// <param name="filename">The filename to render the diagram to</param>
        /// <exception cref="ArgumentNullException">if filename is null</exception>
        /// <exception cref="ArgumentException">if filename (trim'd) is empty</exception>
        internal SVGGraphicsAdapter(string filename)
        {
            if (filename == null)
            {
                throw new ArgumentNullException("filename");
            }
            if (filename.Trim() == string.Empty)
            {
                throw new ArgumentException("filename should not be null.");
            }
            this.filename = filename;
        }

        /// <summary>
        /// Implements the DrawPath interface method and will redirect the DrawPath call to the like name method
        /// in the SVGDrawer.
        /// </summary>
        /// <param name="pen">The pen to draw with</param>
        /// <param name="path">The path to draw</param>
        /// <exception cref="ArgumentNullException">if either arg is null</exception>
        public void DrawPath(Pen pen, GraphicsPath path)
        {
            if (pen == null)
            {
                throw new ArgumentNullException("pen");
            }
            if (path == null)
            {
                throw new ArgumentNullException("path");
            }
            this.svgDrawer.DrawPath(pen, path);
        }

        /// <summary>
        /// Implements the DrawString interface method and will redirect the DrawString call to the like name method
        /// in the SVGDrawer.
        /// </summary>
        /// <param name="text">The text to draw</param>
        /// <param name="font">The font to use</param>
        /// <param name="brush">The brush to draw with</param>
        /// <param name="point">The point to draw it</param>
        /// <exception cref="ArgumentNullException">if any parameter is null</exception>
        public void DrawString(string text, Font font, Brush brush, PointF point)
        {
            if (text == null)
            {
                throw new ArgumentNullException("text");
            }
            if (font == null)
            {
                throw new ArgumentNullException("font");
            }
            if (brush == null)
            {
                throw new ArgumentNullException("brush");
            }
            this.svgDrawer.DrawString(text, font, brush, point);
        }

        /// <summary>
        /// Since the SVG Graphics does not offer any sort of font metrics, we have to generalize this and hope for
        /// the best. This method should:
        /// 1) Create a Bitmap of 1,1 length
        /// 2) Create a graphics from it (Graphics.FromImage(xxx))
        /// 3) Call and return the SizeF from the graphics.MeasureString from step #2
        /// </summary>
        /// <param name="text">The text to measure</param>
        /// <param name="font">The text font</param>
        /// <returns>The size of the text</returns>
        /// <exception cref="ArgumentNullException">if text or font is null</exception>
        public SizeF MeasureString(string text, Font font)
        {
            if (text == null)
            {
                throw new ArgumentNullException("text");
            }
            if (font == null)
            {
                throw new ArgumentNullException("font");
            }
            return Graphics.FromImage(new Bitmap(1, 1)).MeasureString(text, font);
        }

        /// <summary>
        /// This method will save the underlying SVG image. This method will simply call the
        /// svgDrawer.SaveAsSVG(filename, null).  Catch any exception thrown and wrap/throw a render exception
        /// instead.
        /// </summary>
        /// <exception cref="RendererException">if an exception occurs while saving the SVG image</exception>
        public void Complete()
        {
            try
            {
                this.svgDrawer.SaveAsSVG(filename, null);
            }
            catch (Exception exception)
            {
                throw new RendererException("SVG rendering fails.", exception);
            }
        }

        /// <summary>
        /// Disposes the resources used by calling the dispose on the SVGDrawer.
        /// </summary>
        public void Dispose()
        {
            this.svgDrawer.Dispose();
        }

    }

}
