/**
 * TCS Dependency Map Renderer
 *
 * Copyright (c) 2005 TopCoder, Inc. All rights Reserved.
 */
package com.topcoder.rules.dependencymap.view;

import java.util.HashSet;
import java.util.List;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.Set;

/**
 * This is the main class of the application. This class will be used by the application to:
 *
 * 1) Manipulate the layout manager that will be used to layout the dependency diagrams. Only a single layout
 * manager can be active at any given time.
 * 2) Manipulate the renderer(s) used to render the dependency diagrams. More than one renderer can be used.
 * 3) Provide a dependency and have it layed out into a diagram.
 *
 * @author Pops
 * @author TCSDeveloper
 */
public class DependencyRenderer {

    /**
     * The shape factory used by this renderer to produce the shapes used in the rendering process. This variable
     * is set in the constructor, is immutable (the reference only) once set and will be used in the rendere method.
     * This variable has been marked final to enforce it's immutablility (again - reference only).
     */
    private final ShapeFactory shapeFactory;

    /**
     * The variable represents the renderes that will be used to render the diagram. Each element in the list will
     * be of DiagramRenderer type. A linked list is used since random access is not needed and we can conserve space.
     * This value is set during class construction time, is immutable (the reference not the contents) and will be
     * used in the render method. The list content can be accessed via the addDiagramRenderer(),
     * removeDiagramRenderer() and getDiagramRenderers() methods. This variable has been marked final to ensure it's
     * immutability (again reference only).
     */
    private final List renderers = new ArrayList();

    /**
     * The variable represents the layout manager to use when laying the dependency diagram out. This variable will
     * initially be set at object construction time, is mutatable by the setter method and will be accessed in both
     * the getter method and the render method.
     */
    private LayoutManager layoutManager = new FixedGridLayoutManager();

    /**
     * Constructs the DependencyRenderer with a default ShapeFactory. This method will simply call
     * 'this(new ShapeFactory())'.
     */
    public DependencyRenderer() {
        this(new ShapeFactory());
    }

    /**
     * Constructs the DependencyRenderer from the specified shapeFactory. This method will simply save the
     * shapeFactory reference to the like named variable.
     *
     * @param shapeFactory A non-null shapeFactory to use.
     * @throws NullPointerException if shapeFactory is null
     */
    public DependencyRenderer(ShapeFactory shapeFactory) {
        if (shapeFactory == null) {
            throw new NullPointerException("shapeFactory should not be null.");
        }
        this.shapeFactory = shapeFactory;
    }

    /**
     * Adds a new renderer to the list of renderers that will be used to render a dependency diagram.
     *
     * @param renderer the renderer to add to the list of renderers
     * @throws NullPointerException if renderer is null
     */
    public void addDiagramRenderer(DiagramRenderer renderer) {
        if (renderer == null) {
            throw new NullPointerException("renderer should not be null.");
        }
        this.renderers.add(renderer);
    }

    /**
     * This method will remove the specified renderer from the list of renderers. If the renderer was successfully
     * removed, true will be return (false otherwise).
     *
     * @param renderer The renderer to remove
     * @return True if the renderer was successfully removed
     * @throws NullPointerException if renderer is null
     */
    public boolean removeDiagramRenderer(DiagramRenderer renderer) {
        if (renderer == null) {
            throw new NullPointerException("renderer should not be null.");
        }
        return this.renderers.remove(renderer);
    }

    /**
     * Returns an immutable list that will contain all the renderers that will be used by this class.
     *
     * @return An immutable list of renderers
     */
    public List getDiagramRenderers() {
        return Collections.unmodifiableList(this.renderers);
    }

    /**
     * The method will set the layout manager to use when laying out a dependency diagram.
     *
     * @param layoutManager The layout manager to use when laying out a dependency diagram
     * @throws NullPointerException if layoutManager is null
     */
    public void setLayoutManager(LayoutManager layoutManager) {
        if (layoutManager == null) {
            throw new NullPointerException("layoutManager should not be null.");
        }
        this.layoutManager = layoutManager;
    }

    /**
     * This method will return the layout manager to use when laying out a dependency diagram.
     *
     * @return The layout manager being used to layout the dependency diagram
     */
    public LayoutManager getLayoutManager() {
        return this.layoutManager;
    }

    /**
     * This method is the method that will be used by the application to render the dependency diagram. The layout
     * manager will be used to layout the diagram against each renderer (note: because of differences in the font
     * metrics between renderers, the layout must be repeated for each renderer to account for those differences).
     * This method will:
     *
     * 1) Iterate over the renderers and for each renderer ...
     * 2) Call renderer.getGraphics2D(dependency) to get a graphics object for that renderer.
     *    If an RendererException is thrown, simply let it pass through to the caller.
     * 3) Call the layoutManager.layout(dependency, shapeFactory, graphics) to layout the diagram.
     *    If an RendererException is thrown, do step #4 and then throw to the caller.
     * 4) Call graphics.complete() to allow any post processing (saving of files).
     *    If a RendererException is thrown, let it pass through to the caller.
     * 5) Wrap #3 and #4 in a finally clause to call dispose() on the graphics to ensure resources are released.
     * 6) Repeat until all renderers are done
     *
     * @param dependency The dependency that will be rendered
     * @throws NullPointerException if dependency is null
     * @throws RendererException if an exception occurs during rendering
     */
    public void render(Dependency dependency) throws RendererException {
        if (dependency == null) {
            throw new NullPointerException("dependency should not be null.");
        }

        for (Iterator itr = this.renderers.iterator(); itr.hasNext();) {
            DiagramRenderer renderer = (DiagramRenderer) itr.next();
            layoutManager.layout(dependency, shapeFactory, renderer);
        }
    }
    
    /**
     * This method will render all unique nodes within the dependency map
     * - including the top level node.  This method simply calls
     * renderAll(dependency, true).
     * 
     * @param dependency The dependency to render
     * @throws NullPointerException if dependency is null
     * @throws RendererException if an exception occurs during rendering
     */
    public void renderAll(Dependency dependency) throws RendererException {
        this.renderAll(dependency, true);
    }
    
    /**
     * This method will render all unique nodes within the dependency map
     * - including the top level node if renderTopLevel is true.
     * This method will:
     * 1) Create a hashset to hold 'seen' nodes
     * A 'seen' node is a node that has been rendered (as the top level) and
     * is uniquely identified by it's title (dependency.getTitle())
     * 2) If renderTopLevel is true, call 'render(dependency)' and add the
     * dependency to the 'seen' hashset
     * 3) Get the children of the dependents
     * 4) For each child ...
     * a) If the child has already been 'seen' - skip it
     * b) if not, render the child and add it to the 'seen' hashset
     * c) render the child's children (ie repeat step 4 for the children)
     * 
     * 
     * 
     * @param dependeny The dependency to render
     * @param renderTopLevel Whether the top level node is rendered also
     * @throws NullPointerException if dependency is null
     * @throws RendererException if an exception occurs during rendering
     */
    public void renderAll(Dependency dependency, boolean renderTopLevel) throws RendererException {
        if (dependency == null) {
            throw new NullPointerException("dependency should not be null");
        }
        this.renderAll(dependency, new HashSet());
    }
    
    /**
     * This method will render all unique nodes within the dependency map
     * 
     * @param dependeny The dependency to render
     * @param seen the set contains all seen nodes
     * @throws RendererException if an exception occurs during rendering
     */
    public void renderAll(Dependency dependency, Set seen) throws RendererException {
        this.render(dependency);
        List d = dependency.getDependants();
        for (Iterator it = d.iterator(); it.hasNext();) {
            Dependency c = (Dependency) it.next();
            String title = c.getTitle();
            if (!seen.contains(title)) {
                seen.add(title);
                this.renderAll(c, seen);
            }
        }
    }
}
