/**
 * TCS Dependency Map Renderer
 *
 * Copyright (c) 2005 TopCoder, Inc. All rights Reserved.
 */
package com.topcoder.rules.dependencymap.view;

import java.awt.Color;
import java.awt.Font;
import java.awt.Point;
import java.awt.FontMetrics;
import java.awt.Rectangle;
import java.awt.geom.GeneralPath;
import java.awt.geom.AffineTransform;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

import com.topcoder.util.config.ConfigManager;
import com.topcoder.util.config.ConfigManagerException;
import com.topcoder.util.config.UnknownNamespaceException;
import com.topcoder.util.log.Level;
import com.topcoder.util.log.Log;
import com.topcoder.util.log.LogFactory;

/**
 * This class is an implementation of the LayoutManager that will layout the dependency diagram into a fixed
 * sized grid (where all columns are fixed to a single width and all rows are fixed to a single height).
 *
 * @author Pops
 * @author TCSDeveloper
 */
public class FixedGridLayoutManager implements LayoutManager {

    /**
     * This variable represents the extra padding that should appear between columns.
     * This variable is only used *between* columns and will not provide margins
     * around the outer columns.  This variable is set in the constructor, is
     * mutatable by the setter, will be >= 0 and will be used by the layout mechanisms.
     */
    private int colPadding = 10;

    /**
     * This variable represents the extra padding that should appear between rows.
     * This variable is only used *between* rows and will not provide margins
     * around the outer rows.  This variable is set in the constructor, is
     * mutatable by the setter, will be >= 0 and will be used by the layout mechanisms.
     */
    private int rowPadding = 10;

    /**
     * This variable represents the extra padding that should appear around
     * text (top/bottom/left/right).  This variable is set in the constructor, is
     * mutatable by the setter, will be >= 0 and will be used by the layout mechanisms.
     */
    private int textMargin = 10;

    /**
     * This variable represents the text font that should be used.
     * This variable is set in the constructor, is
     * mutatable by the setter, will never be null, and will be used
     * by the layout mechanisms.
     */
    private Font textFont = new Font("SansSerif", Font.PLAIN, 12);

    /**
     * This variable represents the text color that should be used.
     * This variable is set in the constructor, is
     * mutatable by the setter, will never be null, and will be used
     * by the layout mechanisms.
     */
    private Color textColor = Color.BLACK;

    /**
     * This variable represents the shape line color that should be used.
     * This variable is set in the constructor, is
     * mutatable by the setter, will never be null, and will be used
     * by the layout mechanisms.
     */
    private Color shapeLineColor = Color.BLACK;

    /**
     * This variable represents the shape fill color that should be used.
     * This variable is set in the constructor, is
     * mutatable by the setter, will never be null, and will be used
     * by the layout mechanisms.
     */
    private Color shapeFillColor = Color.WHITE;

    /**
     * This variable represents the arrow line color that should be used.
     * This variable is set in the constructor, is
     * mutatable by the setter, will never be null, and will be used
     * by the layout mechanisms.
     */
    private Color arrowLineColor = Color.BLACK;

    /**
     * This variable represents the arrow fill color that should be used.
     * This variable is set in the constructor, is
     * mutatable by the setter, will never be null, and will be used
     * by the layout mechanisms.
     */
    private Color arrowFillColor = Color.WHITE;
    
    private Log log = LogFactory.getLog();

    /**
     * Constructs the fixed grid layout manager using the classname as the default namespace.
     * This constructor simply calls "this(getClass().getName())".
     */
    public FixedGridLayoutManager() {
        this(FixedGridLayoutManager.class.getName());
    }

    /**
     * Constructs the FixedGridLayoutManager from the given namespace.
     * This method simply saves the namespace to the like named variable.
     * 
     * @param namespace The namespace to use
     * @throws NullPointerException if namespace is null
     * @throws IllegalArgumentException if namespace (trim'd) is empty
     */
    public FixedGridLayoutManager(String namespace) {
        if (namespace == null) {
            throw new NullPointerException("namespace should not be null.");
        }
        if (namespace.trim().length() == 0) {
            throw new IllegalArgumentException("namespace should not be empty");
        }
        ConfigManager manager = ConfigManager.getInstance();
        try {
            this.rowPadding = this.getIntProperty(manager, namespace, "rowPadding", 10, 0);
            this.colPadding = this.getIntProperty(manager, namespace, "colPadding", 10, 0);
            this.textMargin = this.getIntProperty(manager, namespace, "textMargin", 10, 0);
            String textFontName = manager.getString(namespace, "textFontName");
            if (textFontName == null) {
                this.log.log(Level.INFO, "Property textFontName is not found, use default value instead.");
                textFontName = "Plain";
            }
            String textFontStyle = manager.getString(namespace, "textFontStyle");
            if (textFontStyle == null) {
                this.log.log(Level.INFO, "Property textFontStyle is not found, use default value instead.");
                textFontStyle = "SansSerif";
            }
            int style;
            if (textFontStyle.equals("Plain")) {
                style = Font.PLAIN;
            } else if (textFontStyle.equals("Bold")) {
                style = Font.BOLD;
            } else if (textFontStyle.equals("Italic")) {
                style = Font.ITALIC;
            } else if (textFontStyle.equals("Bold/Italic")) {
                style = Font.BOLD | Font.ITALIC;
            } else {
                this.log.log(Level.INFO, "Property textFontStyle is invalid, use default value instead.");
                style = Font.PLAIN;
            }
            int textFontSize = this.getIntProperty(manager, namespace, "textFontSize", 12, 8);
            this.textFont = new Font(textFontName, style, textFontSize);
            this.textColor = this.getColorProperty(manager, namespace, "textColor", Color.BLACK);
            this.shapeLineColor = this.getColorProperty(manager, namespace, "shapeLineColor", Color.BLACK);
            this.shapeFillColor = this.getColorProperty(manager, namespace, "shapeFillColor", Color.WHITE);
            this.arrowLineColor = this.getColorProperty(manager, namespace, "arrowLineColor", Color.BLACK);
            this.arrowFillColor = this.getColorProperty(manager, namespace, "arrowFillColor", Color.WHITE);
        } catch (UnknownNamespaceException e) {
            this.log.log(Level.ERROR, e.getMessage());
        }
    }
    
    /**
     * Returns the int value of the property specified from ConfigManager. If property is not found or invalid, 
     * the default value will be returned.
     * 
     * @param manager the ConfigManager to use
     * @param namespace the namespace of property
     * @param prop the name of property
     * @param defaultValue the default value
     * @param lowLimit the minimum value of property
     * @return the int value of specified property
     * @throws ConfigManagerException if specified namespace not found
     */
    private int getIntProperty(ConfigManager manager, String namespace, String prop, int defaultValue, int lowLimit)
        throws UnknownNamespaceException {
        String s = manager.getString(namespace, prop);
        if (s == null) {
            // property not found
            this.log.log(Level.INFO, "Property " + prop + " is not found, use default value instead.");
            return defaultValue;
        }
        try {
            int ret = Integer.parseInt(s);
            if (ret < lowLimit) {
                // out of range
                this.log.log(Level.INFO, "Property " + prop + " should be greater than "
                        + lowLimit + ", use default value instead.");
                return defaultValue;
            }
            return ret;
        } catch (NumberFormatException e) {
            // invalid value
            this.log.log(Level.INFO, "Property " + prop + " is not an integer, use default value instead.");
            return defaultValue;
        }
    }
    
    /**
     * Returns the Color with RGB value in the specified property. If property is not found or invalid, 
     * the default value will be returned.
     * 
     * @param manager the ConfigManager to use
     * @param namespace the namespace of property
     * @param prop the name of property
     * @param defaultValue the default color
     * @return the Color with RGB value in the specified property
     * @throws ConfigManagerException if specified namespace not found
     */
    private Color getColorProperty(ConfigManager manager, String namespace, String prop, Color defaultValue)
        throws UnknownNamespaceException {
        return new Color(this.getIntProperty(manager, namespace, prop, defaultValue.getRGB(), Integer.MIN_VALUE));
    }    
    
    /**
     * Calculate the widths of each of the unseen dependencies with is referenced by a root dependency.
     *
     * @param dependency the root dependency
     * @return a mapping from dependency to widths (Integer)
     */
    private Map getWidths(Dependency dependency) {
        // Mapping from dependants to widths (number of cells)
        Map widths = new HashMap();
        // Visitation count for dependants (Integer)
        Map seen = new HashMap();
        // BFS queue
        List queue = new ArrayList();

        // Initialize the queue and visitation count by placing the root dependency.
        queue.add(dependency);
        seen.put(dependency, new Integer(1));

        // Iterate through the queue.
        for (int i = 0; i < queue.size(); ++i) {
            dependency = (Dependency) queue.get(i);

            for (Iterator itr = dependency.getDependants().iterator(); itr.hasNext();) {
                Dependency dependant = (Dependency) itr.next();

                // If the dependency is seen for the first time, we need to calculate its
                // width, otherwise it's a seen dependency and has constant width of 1.
                if (!seen.containsKey(dependant)) {
                    queue.add(dependant);
                    seen.put(dependant, new Integer(1));
                } else {
                    seen.put(dependant, new Integer(((Integer) seen.get(dependant)).intValue() + 1));
                }
            }
        }

        // Calculate the width in a backward order.
        for (int i = queue.size() - 1; i >= 0; --i) {
            dependency = (Dependency) queue.get(i);

            int width = 0;
            boolean allOne = true;
            
            for (Iterator itr = dependency.getDependants().iterator(); itr.hasNext();) {
                Dependency dependant = (Dependency) itr.next();

                // If the dependency is an unseen dependency, its width should already been calculated.
                // Otherwise its width is constantly 1.
                if (seen.get(dependant).equals(new Integer(1))) {
                    int w = ((Integer) widths.get(dependant)).intValue();
                    width += w;
                    if (w > 1) {
                        allOne = false;
                    }
                } else {
                    ++width;
                    seen.put(dependant, new Integer(((Integer) seen.get(dependant)).intValue() - 1));
                }
            }
            
            // if even and all chidren's width is 1
            if (width % 2 == 0 && allOne) {
                ++width;
            }

            // Store the width into the mapping.
            widths.put(dependency, new Integer(width));
        }

        return widths;
    }

    /**
     * Paint a dependency.
     *
     * @param dependency the dependency to paint
     * @param shape the shape for outbound
     * @param width the width of a cell
     * @param height the height of a cell
     * @param arrowHeight the height of an arrow
     * @param x the x coord position
     * @param y the y coord position
     * @param graphics the graphics to paint onto
     */
    private void paintDependency(Dependency dependency, GeneralPath shape, int width, int height,
                                 int arrowHeight, float x, int y, RendererGraphics graphics) {
        // Calculate pixel offset.
        x *= width + this.colPadding;
        y *= height + arrowHeight + this.rowPadding;

        // Scale it to cell size and translate it to position.
        Rectangle rect = shape.getBounds();
        shape.transform(AffineTransform.getScaleInstance(((double) width) / rect.width,
                                                         ((double) height) / rect.height));
        shape.transform(AffineTransform.getTranslateInstance(x, y));

        // Draw the shape.
        graphics.setColor(this.shapeFillColor);
        graphics.fill(shape);
        graphics.setColor(this.shapeLineColor);
        graphics.draw(shape);

        // Retrieve font metrics and estimate text width.
        FontMetrics metrics = graphics.getFontMetrics(graphics.getFont());
        int myTextWidth = metrics.stringWidth(dependency.getTitle());

        // Draw the text to the center of the cell.
        graphics.setColor(this.textColor);
        graphics.drawString(dependency.getTitle(), (int) x + this.textMargin +
                            (width - this.textMargin * 2 - myTextWidth) / 2,
                                                   y + this.textMargin + 1 + metrics.getAscent());
    }

    /**
     * Paint an arrow.
     *
     * @param shape the shape for the arrow
     * @param x1 the x coord of the first dependency
     * @param x2 the x coord of the second dependency
     * @param y the y coord of the first dependency
     * @param width the width of a cell
     * @param height the height of a cell
     * @param graphics the graphics to paint onto
     */
    private void paintArrow(GeneralPath shape, float x1, int x2, int y,
                            int width, int height, RendererGraphics graphics) {
        Rectangle rect = shape.getBounds();
        // Calculate arrow position.
        float x = (float) ((x1 + x2) / 2.0);
        x *= width + this.colPadding;
        x += (width - rect.width) / 2;
        y *= height + rect.height + this.rowPadding;
        y += height + this.rowPadding / 2;

        // Rotate it to point the dependant and translate it to position.
        shape.transform(AffineTransform.getRotateInstance(
                Math.PI / 2 - Math.atan2(height + rect.height + this.rowPadding,
                                         (x1 - x2) * (width + this.colPadding)),
                rect.width / 2.0, rect.height / 2.0));
        shape.transform(AffineTransform.getTranslateInstance(x, y));

        // Draw the arrow.
        graphics.setColor(this.arrowFillColor);
        graphics.fill(shape);
        graphics.setColor(this.arrowLineColor);
        graphics.draw(shape);
    }

    /**
     * Paint the dependency hiearchy.
     *
     * @param dependency the root dependency
     * @param widths the mapping of the width of the dependencies
     * @param shapeFactory the shape factory to produce the shapes
     * @param graphics the graphics to paint onto
     * @throws RendererException
     */
    private void paint(Dependency dependency, Map widths, ShapeFactory shapeFactory, DiagramRenderer renderer)
        throws RendererException {
        // get the graphics to measure string sizes 
        RendererGraphics graphics = renderer.getGraphics(dependency, 1, 1);

        // Calculate the maximum text size.
        FontMetrics metrics = graphics.getFontMetrics(graphics.getFont());
        int textHeight = metrics.getHeight();
        int textWidth = 0;
        for (Iterator itr = widths.keySet().iterator(); itr.hasNext();) {
            int myTextWidth = metrics.stringWidth(((Dependency) itr.next()).getTitle());

            if (myTextWidth > textWidth) {
                textWidth = myTextWidth;
            }
        }
        int height = textHeight + this.textMargin * 2 + 2;
        int width = textWidth + this.textMargin * 2 + 2;
        
        graphics.dispose();
        // re-get the graphics;
        graphics = renderer.getGraphics(dependency, height, width);
        
        try {
            int arrowHeight = shapeFactory.getShape(ShapeFactory.RELATIONSHIP_ARROW).getBounds().height;

            // Visitation set
            Set seen = new HashSet();
            // BFS queue
            List queue = new ArrayList();
            // Position of the dependency to render
            List positions = new ArrayList();
            // The position of the parent dependency
            List parentPos = new ArrayList();
    
            // Initialize by inserting the root dependency.
            seen.add(dependency);
            queue.add(dependency);
            positions.add(new Point(0, 0));
            parentPos.add(null);
    
            // Render through the queue.
            for (int i = 0; i < queue.size(); ++i) {
                dependency = (Dependency) queue.get(i);
                Point position = (Point) positions.get(i);
    
                // Calculate the dependants width to decide whether we need a padding cell.
                int dependantsWidth = 0;
                boolean allOne = true;
                for (Iterator itr = dependency.getDependants().iterator(); itr.hasNext();) {
                    Dependency dependant = (Dependency) itr.next();
    
                    if (!seen.contains(dependant)) {
                        int w = ((Integer) widths.get(dependant)).intValue();
                        if (w > 1) {
                            allOne = false;
                        }
                        dependantsWidth += w;
                    } else {
                        ++dependantsWidth;
                    }
                }
                boolean padding = dependantsWidth % 2 == 0 && allOne;
       
                // Obtain shape used for this unseen shape.
                GeneralPath shape = null;
                if (dependency.getShapeOverride() != -1) {
                    shape = shapeFactory.getShape(dependency.getShapeOverride());
                }
                if (shape == null) {
                    shape = shapeFactory.getShape(ShapeFactory.UNSEEN_ELEMENT);
                }
                
                float dx;
                
                int w = ((Integer) widths.get(dependency)).intValue();
                //dependency.title = "" + w + " " + dependantsWidth;
                if (parentPos.get(i) == null) {
                    // top level
                    if (dependency.getDependants().size() == 1) {
                        dx = position.x + (w - 1) / 2;
                    } else {
                        dx = position.x + (w - 1) / 2.0f;
                    }
                } else if (w % 2 == 0) {
                    float parent = ((Float) parentPos.get(i)).floatValue();
                    dx = position.x + (w - 1) / 2.0f;
                    if (dx < parent) {
                        dx += 0.5f;
                    } else {
                        dx -= 0.5f;
                    }
                } else {
                    dx = position.x + (w - 1) / 2;
                }
                paintDependency(dependency, shape, width, height,
                        arrowHeight, dx, position.y, graphics);
                
                // If parent dependency exists we make the relationship arrow.
                if (parentPos.get(i) != null) {
                    float parent = ((Float) parentPos.get(i)).floatValue();
                    paintArrow(shapeFactory.getShape(ShapeFactory.RELATIONSHIP_ARROW),
                               parent, (int) dx, position.y - 1, width, height, graphics);
                }

                // Accumulate the width to find the position for each dependant.
                int cumulativeWidth = 0;
                for (Iterator itr = dependency.getDependants().iterator(); itr.hasNext();) {
                    Dependency dependant = (Dependency) itr.next();
    
                    // Check to see if we make the padding on this dependant.
                    if (padding && cumulativeWidth * 2 >= dependantsWidth) {
                        padding = false;
                        ++cumulativeWidth;
                    }
    
                    if (!seen.contains(dependant)) {
    
                        // Push the unseen dependant into the queue.
                        seen.add(dependant);
                        queue.add(dependant);
                        positions.add(new Point(position.x + cumulativeWidth, position.y + 1));
                        parentPos.add(new Float(dx));
    
                        cumulativeWidth += ((Integer) widths.get(dependant)).intValue();
                    } else {
                        // Draw arrow for the seen dependant.
                        paintArrow(shapeFactory.getShape(ShapeFactory.RELATIONSHIP_ARROW),
                                   position.x + dependantsWidth / 2, position.x + cumulativeWidth, position.y,
                                   width, height, graphics);
    
                        // Obtain shape used for this seen shape.
                        shape = null;
                        if (dependant.getShapeOverride() != -1) {
                            shape = shapeFactory.getShape(dependant.getShapeOverride());
                        }
                        if (shape == null) {
                            shape = shapeFactory.getShape(ShapeFactory.SEEN_ELEMENT);
                        }
    
                        // Draw the dependant.
                        paintDependency(dependant, shape, width, height,
                                        arrowHeight, position.x + cumulativeWidth, position.y + 1, graphics);
    
                        ++cumulativeWidth;
                    }
                }
            }
            graphics.complete();
        } finally {
            graphics.dispose();
        }
    }

    /**
     * Lays out and prints the dependency diagram. Since the algorithm to do this is beyond Poseidon's
     * capabilities - please review the component specifications for details.
     *
     * @param dependency The dependency to layout
     * @param shapeFactory The shape factory to use
     * @param renderer The renderer that will be used
     * @throws NullPointerException if any parameter is null
     * @throws RendererException if some exception occurred while laying out the dependency diagram
     */
    public void layout(Dependency dependency, ShapeFactory shapeFactory, DiagramRenderer renderer)
        throws RendererException {
        if (dependency == null) {
            throw new NullPointerException("dependency should not be null.");
        }
        if (shapeFactory == null) {
            throw new NullPointerException("shapeFactory should not be null.");
        }
        if (renderer == null) {
            throw new NullPointerException("graphics should not be null.");
        }

        Map widths = getWidths(dependency);

        paint(dependency, widths, shapeFactory, renderer);
    }

    /**
     * The getter of arrowFillColor
     * 
     * @return the arrowFillColor.
     */
    public Color getArrowFillColor() {
        return arrowFillColor;
    }
    
    /**
     * The setter of arrowFillColor
     * 
     * @param arrowFillColor The arrowFillColor to set.
     */
    public void setArrowFillColor(Color arrowFillColor) {
        this.arrowFillColor = arrowFillColor;
    }
    
    /**
     * The getter of arrowLineColor
     * 
     * @return the arrowLineColor.
     */
    public Color getArrowLineColor() {
        return arrowLineColor;
    }
    
    /**
     * The setter of arrowLineColor
     * 
     * @param arrowLineColor The arrowLineColor to set.
     */
    public void setArrowLineColor(Color arrowLineColor) {
        this.arrowLineColor = arrowLineColor;
    }
    
    /**
     * The getter of colPadding
     * 
     * @return the colPadding.
     */
    public int getColPadding() {
        return colPadding;
    }
    
    /**
     * The setter of colPadding
     * 
     * @param colPadding The colPadding to set.
     */
    public void setColPadding(int colPadding) {
        this.colPadding = colPadding;
    }
    
    /**
     * The getter of rowPadding
     * 
     * @return the rowPadding.
     */
    public int getRowPadding() {
        return rowPadding;
    }
    
    /**
     * The setter of rowPadding
     * 
     * @param rowPadding The rowPadding to set.
     */
    public void setRowPadding(int rowPadding) {
        this.rowPadding = rowPadding;
    }
    
    /**
     * The getter of shapeFillColor
     * 
     * @return the shapeFillColor.
     */
    public Color getShapeFillColor() {
        return shapeFillColor;
    }
    
    /**
     * The setter of shapeFillColor
     * 
     * @param shapeFillColor The shapeFillColor to set.
     */
    public void setShapeFillColor(Color shapeFillColor) {
        this.shapeFillColor = shapeFillColor;
    }
    
    /**
     * The getter of shapeLineColor
     * 
     * @return the shapeLineColor.
     */
    public Color getShapeLineColor() {
        return shapeLineColor;
    }
    
    /**
     * The setter of shapeLineColor
     * 
     * @param shapeLineColor The shapeLineColor to set.
     */
    public void setShapeLineColor(Color shapeLineColor) {
        this.shapeLineColor = shapeLineColor;
    }
    
    /**
     * The getter of textColor
     * 
     * @return the textColor.
     */
    public Color getTextColor() {
        return textColor;
    }
    
    /**
     * The setter of textColor
     * 
     * @param textColor The textColor to set.
     */
    public void setTextColor(Color textColor) {
        this.textColor = textColor;
    }
    
    /**
     * The getter of textFont
     * 
     * @return the textFont.
     */
    public Font getTextFont() {
        return textFont;
    }
    
    /**
     * The setter of textFont
     * 
     * @param textFont The textFont to set.
     */
    public void setTextFont(Font textFont) {
        this.textFont = textFont;
    }
    
    /**
     * The getter of textMargin
     * 
     * @return the textMargin.
     */
    public int getTextMargin() {
        return textMargin;
    }
    
    /**
     * The setter of textMargin
     * 
     * @param textMargin The textMargin to set.
     */
    public void setTextMargin(int textMargin) {
        this.textMargin = textMargin;
    }
}
